--
-- FS19 - Mountable attacher
-- @author:    	kenny456 (kenny456@seznam.cz)
-- @history:	v1.0.0.0 - 2020-12-27 - first release
--
MountableAttacher = {}
MountableAttacher.confDir = getUserProfileAppPath().. "modsSettings/MountableAttacher/"
MountableAttacher.modDirectory = g_currentModDirectory
MountableAttacher.showHelp = true
MountableAttacher.mountedVehicles = {}

function MountableAttacher.prerequisitesPresent(specializations)
	return true
end
function MountableAttacher.registerOverwrittenFunctions(vehicleType)
	if MountableAttacher.maPlayer == nil then
		Player.registerActionEvents = Utils.appendedFunction(Player.registerActionEvents, MountableAttacher.registerActionEventsPlayer)
		MountableAttacher.maPlayer = true
	end
end
function MountableAttacher.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "mountAttacher", 						MountableAttacher.mountAttacher)
	SpecializationUtil.registerFunction(vehicleType, "unmountAttacher", 					MountableAttacher.unmountAttacher)
	SpecializationUtil.registerFunction(vehicleType, "moveAttacher", 						MountableAttacher.moveAttacher)
	SpecializationUtil.registerFunction(vehicleType, "playerInRangeMA", 					MountableAttacher.playerInRangeMA)
	SpecializationUtil.registerFunction(vehicleType, "preMountAttacher", 					MountableAttacher.preMountAttacher)
	SpecializationUtil.registerFunction(vehicleType, "updateDirtAndWear", 					MountableAttacher.updateDirtAndWear)
	SpecializationUtil.registerFunction(vehicleType, "maGetFullName", 						MountableAttacher.maGetFullName)
end
function MountableAttacher:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function MountableAttacher:registerActionEventsPlayer()
	local spec = self.spec_mountableAttacher
	
	if g_dedicatedServerInfo ~= nil then
		return
	end
	if MountableAttacher.event_IDsPlayer == nil then
		MountableAttacher.event_IDsPlayer = {}
	end
	local actions = { InputAction.MOUNTABLE_ATTACHER_MOUNT, InputAction.MOUNTABLE_ATTACHER_MOVE_TRIGGER, InputAction.MOUNTABLE_ATTACHER_MOVE_LEFT, InputAction.MOUNTABLE_ATTACHER_MOVE_RIGHT,
					InputAction.MOUNTABLE_ATTACHER_MOVE_FORWARD, InputAction.MOUNTABLE_ATTACHER_MOVE_BACKWARD, InputAction.MOUNTABLE_ATTACHER_MOVE_UP, InputAction.MOUNTABLE_ATTACHER_MOVE_DOWN }

	for _,actionName in pairs(actions) do
		local always = (actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_LEFT or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_RIGHT or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_FORWARD or
						actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_BACKWARD or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_UP or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_DOWN or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_TRIGGER) and true or false
		local __, eventID = InputBinding.registerActionEvent(g_inputBinding, actionName, self, MountableAttacher.actionCallbackPlayer, true, true, always, true)
		MountableAttacher.event_IDsPlayer[actionName] = eventID
		if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
			g_inputBinding:setActionEventTextVisibility(eventID, MountableAttacher.showHelp)
			g_inputBinding:setActionEventActive(eventID, false)
		end
		local colliding = false
		_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
		if colliding then
			if g_inputBinding.nameActions[actionName].bindings[1] ~= nil then
				if g_inputBinding.nameActions[actionName].bindings[1].inputString ~= nil then
					print(string.format('Warning: MountableAttacher got a colliding input action: %s', actionName)..' ('..g_inputBinding.nameActions[actionName].bindings[1].inputString..'). You can remap it in controls settings')
				end
			else
				print(string.format('Warning: MountableAttacher got a colliding input action: %s', actionName))
			end
		end
	end
end
function MountableAttacher.registerEventListeners(vehicleType)
	for _,n in pairs( { "onLoad", "onPostLoad", "onDelete", "saveToXMLFile", "onUpdate", "onUpdateTick", "onDraw", "onReadStream", "onWriteStream", "onRegisterActionEvents", "mountAttacher", "unmountAttacher", "moveAttacher" } ) do
		SpecializationUtil.registerEventListener(vehicleType, n, MountableAttacher)
	end
end
function MountableAttacher:onLoad(savegame)
	self.spec_mountableAttacher = {}
	local spec = self.spec_mountableAttacher
	
	spec.modInitialized = false
	spec.modAllowed = true
	if self.spec_attacherJoints.attacherJoints ~= nil then
		spec.modInitialized = true
	end
	if not spec.modInitialized then
		print("Error: MountableAttacher initialization failed for "..tostring(self:maGetFullName()).." ! ")
		return
	else
		--print("----------------------------------MountableAttacher initialized for "..tostring(self:maGetFullName()).." ! ")
	end
	MountableAttacher.vehicles = {}
	spec.attacherMounted = false
	spec.attacherMountedBack = false
	spec.timer = 0
	spec.attacherImplementsVisible = false
end
function MountableAttacher:onPostLoad(savegame)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if savegame ~= nil then
		local xmlFile = savegame.xmlFile
		local key = savegame.key ..".MountableAttacher"
		local attacherMounted = Utils.getNoNil(getXMLBool(xmlFile, key.."#attacherMounted"), spec.attacherMounted)
		local attacherMountedBack = Utils.getNoNil(getXMLBool(xmlFile, key.."#attacherMountedBack"), spec.attacherMountedBack)
		spec.attacherMountedBack = attacherMountedBack
		if attacherMounted then
			spec.hookXPost = getXMLFloat(xmlFile, key.."#hookX")
			spec.hookYPost = getXMLFloat(xmlFile, key.."#hookY")
			spec.hookZPost = getXMLFloat(xmlFile, key.."#hookZ")
			spec.preMount = true
			self.loadAttachmentsFromXMLFile = Utils.prependedFunction(self.loadAttachmentsFromXMLFile, self.preMountAttacher)
		end
	end
end
function MountableAttacher:preMountAttacher()
	local spec = self.spec_mountableAttacher
	
	self:mountAttacher(spec.attacherMountedBack, true)
	self:moveAttacher(spec.hookXPost, spec.hookYPost, spec.hookZPost, true)
	spec.preMounted = true
end
function MountableAttacher:onDelete()
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	MountableAttacher.vehicles[self] = nil
	MountableAttacher.mountedVehicles[self] = nil
end
function MountableAttacher:saveToXMLFile(xmlFile, key)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	setXMLBool(xmlFile, key.."#attacherMounted", spec.attacherMounted)
	if spec.attacherMounted then
		setXMLBool(xmlFile, key.."#attacherMountedBack", spec.attacherMountedBack)
		setXMLFloat(xmlFile, key.."#hookX", spec.hookX)
		setXMLFloat(xmlFile, key.."#hookY", spec.hookY)
		setXMLFloat(xmlFile, key.."#hookZ", spec.hookZ)
	end
end
function MountableAttacher:onUpdate(dt)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if spec.preMount and not spec.preMounted then
		spec.preMounted = true
		self:mountAttacher(spec.attacherMountedBack, false)
		self:moveAttacher(spec.hookXPost, spec.hookYPost, spec.hookZPost, false)
	end
	self:raiseActive()
	local vehicleInRange = false
	local canBeControlled = false
	local anyVehicleInRange = false
	local count = 0
	if self.isClient then
		if spec.attacherMounted then
			if self:getImplementByJointDescIndex(spec.mountedAttacherIndex) ~= nil then
				if not spec.attacherImplementsVisible then
					spec.attacherImplementsVisible = true
					setVisibility(spec.attacherImplements, true)
				end
			elseif self:getImplementByJointDescIndex(spec.mountedAttacherIndex) == nil then
				if spec.attacherImplementsVisible then
					spec.attacherImplementsVisible = false
					setVisibility(spec.attacherImplements, false)
				end
			end
		end
		if not self:getIsActiveForInput() and g_currentMission.controlPlayer and g_currentMission.player ~= nil then
			local x,y,z = getWorldTranslation(g_currentMission.player.pickUpKinematicHelperNode)
			local vx, vy, vz = getWorldTranslation(self.rootNode)
			local distance = MathUtil.vector2Length(x - vx, z - vz)
			if distance <= self.sizeLength / 1.8 then
				vehicleInRange = true
				if MountableAttacher.vehicles[self] == nil then
					MountableAttacher.vehicles[self] = {}
					MountableAttacher.vehicles[self].id = self
					MountableAttacher.vehicles[self].distance = distance
				else
					MountableAttacher.vehicles[self].distance = distance
				end
			else
				if MountableAttacher.vehicles[self] ~=  nil then
					MountableAttacher.vehicles[self] = nil
				end
			end
			for _, vehicle in pairs(MountableAttacher.vehicles) do
				count = count + 1
			end
			anyVehicleInRange = count > 0
			if anyVehicleInRange then
				if MountableAttacher.vehicles[self] ~=  nil then
					local nearest
					local distanceTemp = 10000
					for _, vehicle in pairs(MountableAttacher.vehicles) do
						if vehicle.distance < distanceTemp then
							distanceTemp = vehicle.distance
							nearest = vehicle.id
						end
					end
					if nearest == self then
						canBeControlled = true
						MountableAttacher.controlledVehicle = self
					end
				end
			else
				MountableAttacher.controlledVehicle = nil
			end
			if canBeControlled then
				if MountableAttacher.input_mountAttacher then
					MountableAttacher.input_mountAttacher = false
					if not spec.attacherMounted then
						local x,y,z = getWorldTranslation(g_currentMission.player.rootNode)
						local xx,yy,zz = worldToLocal(self.rootNode, x,y,z)
						attacherMountedBack = zz < 0
						self:mountAttacher(attacherMountedBack, false)
					else
						if self:getImplementByJointDescIndex(spec.mountedAttacherIndex) == nil and self:getImplementByJointDescIndex(spec.mountedAttacherIndex-1) == nil and self:getImplementByJointDescIndex(spec.mountedAttacherIndex-2) == nil then
							self:unmountAttacher(false)
						else
							g_currentMission:showBlinkingWarning('First detach implement to unmout this attacher.', 2000)
						end
					end
				end
			end
		else
			if MountableAttacher.vehicles[self] ~=  nil then
				MountableAttacher.vehicles[self] = nil
			end
			MountableAttacher.controlledVehicle = nil
		end
	end
	if self.isClient then
		if MountableAttacher.event_IDsPlayer ~= nil then
			for actionName,eventID in pairs(MountableAttacher.event_IDsPlayer) do
				if anyVehicleInRange then
					if canBeControlled then
						if actionName == InputAction.MOUNTABLE_ATTACHER_MOUNT then
							g_inputBinding:setActionEventActive(eventID, vehicleInRange)
							g_inputBinding:setActionEventText(eventID, spec.attacherMounted and g_i18n:getText('MOUNTABLE_ATTACHER_UNMOUNT') ..' ' .. self:maGetFullName() or g_i18n:getText('MOUNTABLE_ATTACHER_MOUNT') ..' ' .. self:maGetFullName())
							g_inputBinding:setActionEventTextVisibility(eventID, true)
						elseif actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_TRIGGER then
							g_inputBinding:setActionEventActive(eventID, spec.attacher ~= nil and MountableAttacher.playerInRangeMA(self, spec.attacher, 4) and spec.attacherMounted and self:getImplementByJointDescIndex(spec.mountedAttacherIndex) == nil and self:getImplementByJointDescIndex(spec.mountedAttacherIndex-1) == nil and self:getImplementByJointDescIndex(spec.mountedAttacherIndex-2) == nil)
							g_inputBinding:setActionEventTextVisibility(eventID, not spec.input_moveTrigger)
						elseif actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_LEFT or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_RIGHT or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_FORWARD or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_BACKWARD
						or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_UP or actionName == InputAction.MOUNTABLE_ATTACHER_MOVE_DOWN then
							g_inputBinding:setActionEventActive(eventID, spec.attacher ~= nil and MountableAttacher.playerInRangeMA(self, spec.attacher, 4) and spec.input_moveTrigger)
							g_inputBinding:setActionEventTextVisibility(eventID, true)
						end
					end
				else
					g_inputBinding:setActionEventActive(eventID, false)
				end
			end
		end
		if spec.attacher ~= nil and canBeControlled then
			if MountableAttacher.playerInRangeMA(self, spec.attacher, 4) then
				if spec.input_moveLeft or spec.input_moveRight or spec.input_moveForward or spec.input_moveBackward or spec.input_moveUp or spec.input_moveDown then
					spec.hookIsMoving = true
					local x, y, z = getTranslation(spec.attacher)
					local origZ = self.spec_towingChain and -0.1 or self.sizeLength / 2.3 * (spec.attacherMountedBack and -1 or 1)
					spec.timerHook = spec.timerHook + dt
					if spec.input_moveLeft then
						if spec.timerHook > 1000 then
							x = x + 0.01
						else
							x = x + 0.001
						end
						if x > 1.5 then x = 1.5 end
					end
					if spec.input_moveRight then
						if spec.timerHook > 1000 then
							x = x - 0.01
						else
							x = x - 0.001
						end
						if x < -1.5 then x = -1.5 end
					end
					if spec.input_moveForward then
						if spec.timerHook > 1000 then
							z = z + 0.01
						else
							z = z + 0.001
						end
						if z > origZ + 2 then z = origZ + 2 end
					end
					if spec.input_moveBackward then
						if spec.timerHook > 1000 then
							z = z - 0.01
						else
							z = z - 0.001
						end
						if z < origZ - 2 then z = origZ - 2 end
					end
					if spec.input_moveUp then
						if spec.timerHook > 1000 then
							y = y + 0.01
						else
							y = y + 0.001
						end
						if y > 2.5 then y = 2.5 end
					end
					if spec.input_moveDown then
						if spec.timerHook > 1000 then
							y = y - 0.01
						else
							y = y - 0.001
						end
						if y < 0 then y = 0 end
					end
					spec.hookX = x
					spec.hookY = y
					spec.hookZ = z
					setTranslation(spec.attacher, x, y, z)
				else
					spec.timerHook = 0
				end
			else
				if spec.hookIsMoving then
					spec.hookIsMoving = false
					spec.input_moveTrigger = false
					spec.input_moveLeft = false
					spec.input_moveRight = false
					spec.input_moveForward = false
					spec.input_moveBackward = false
					spec.input_moveUp = false
					spec.input_moveDown = false
					self:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false)
				end
			end
		end
	end
end
function MountableAttacher:onUpdateTick(dt)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if spec.attacherMounted and self.spec_washable ~= nil and self.spec_washable.washableNodes ~= nil then
		local x, y, z, w = getShaderParameter(self.spec_washable.washableNodes[1].nodes[1], "RDT")
		if x ~= spec.wearLast or y ~= spec.dirtLast then
			spec.wearLast = x
			spec.dirtLast = y
			self:updateDirtAndWear()
		end
	end
end
function MountableAttacher:mountAttacher(attacherMountedBack, noEventSend)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	MountableAttacherMountEvent.sendEvent(self, attacherMountedBack, noEventSend)
	
	spec.attacherI3d = g_i3DManager:loadSharedI3DFile('attacher.i3d', MountableAttacher.modDirectory, false, false)
	spec.attacher = I3DUtil.indexToObject(spec.attacherI3d, '0|0')
	spec.attacherJoint = I3DUtil.indexToObject(spec.attacherI3d, '0|0|0|0')
	spec.attacherImplements = I3DUtil.indexToObject(spec.attacherI3d, '0|0|0|2')
	spec.washIndexes = {'0|0|0', '0|0|0|1', '0|0|0|2|0', '0|0|0|2|1', '0|0|0|2|2', '0|0|0|2|3', '0|0|0|2|4'}
	spec.washNodes = {}
	for i, node in pairs(spec.washIndexes) do
		spec.washNodes[i] = I3DUtil.indexToObject(spec.attacherI3d, node)
	end
	setVisibility(spec.attacherImplements, false)
	local attacherFile = MountableAttacher.modDirectory.."attacher.xml"
	spec.attacherXml = loadXMLFile("MountableAttacher_xml", attacherFile)
	local i=0
	while true do
		local baseName = string.format("vehicle.attacherJoints.attacherJoint(%d)", i)
		if not hasXMLProperty(spec.attacherXml, baseName) then
			break
		end
		local attacherJoint = {}
		if self:loadAttacherJointFromXML(attacherJoint, spec.attacherXml, baseName, i) then
			table.insert(self.spec_attacherJoints.attacherJoints, attacherJoint)
			attacherJoint.index = #self.spec_attacherJoints.attacherJoints
			attacherJoint.jointTransform  = spec.attacherJoint
			spec.mountedAttacherIndex = attacherJoint.index
			attacherJoint.jointOrigRot = { getRotation(attacherJoint.jointTransform) }
			attacherJoint.jointOrigTrans = { getTranslation(attacherJoint.jointTransform) }
		end
		i = i + 1
	end
	spec.attacherMountedBack = attacherMountedBack
	spec.hookX = 0
	spec.hookY = self.spec_towingChain ~= nil and 0 or 1
	spec.hookZ = self.spec_towingChain ~= nil and 0 or ((self.sizeLength/2.3) * (spec.attacherMountedBack and -1 or 1))
	setTranslation(spec.attacher, spec.hookX, spec.hookY, spec.hookZ)
	setRotation(spec.attacher, 0, self.spec_towingChain ~= nil and 0 or (spec.attacherMountedBack and math.rad(0) or math.rad(180)), 0)
	link(self.rootNode, spec.attacherI3d)
	spec.attacherMounted = true
	self:updateDirtAndWear()
end
function MountableAttacher:unmountAttacher(noEventSend)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	MountableAttacherUnmountEvent.sendEvent(self, noEventSend)
	spec.attacherMounted = false
	self.spec_attacherJoints.attacherJoints[spec.mountedAttacherIndex] = nil
	self.spec_attacherJoints.attacherJoints[spec.mountedAttacherIndex-1] = nil
	self.spec_attacherJoints.attacherJoints[spec.mountedAttacherIndex-2] = nil
	g_i3DManager:releaseSharedI3DFile('hook.i3d', nil, true)
	if self.isServer then
		if entityExists(self.rootNode) then
			delete(spec.attacher)
		end
	elseif self.isClient then
		if self ~= nil then
			delete(spec.attacher)
		end
	end
	spec.attacher = nil
	spec.attacherJoint  = nil
	spec.attacherImplements  = nil
	spec.attacherMountedBack = false
end
function MountableAttacher:moveAttacher(x, y, z, noEventSend)
	local spec = self.spec_mountableAttacher
	MountableAttacherMoveEvent.sendEvent(self, x, y, z, noEventSend)
	
	setTranslation(spec.attacher, x, y, z)
	spec.hookX = x
	spec.hookY = y
	spec.hookZ = z
end
function MountableAttacher:updateDirtAndWear()
	local spec = self.spec_mountableAttacher
	
	if self.spec_washable ~= nil and self.spec_washable.washableNodes ~= nil then
		local x, y, z, w = getShaderParameter(self.spec_washable.washableNodes[1].nodes[1], "RDT")
		for _, node in pairs(spec.washNodes) do
			setShaderParameter(node, "RDT", x, y, z, w, false)
		end
	end
end
function MountableAttacher:actionCallbackPlayer(actionName, keyStatus, arg4, arg5, arg6)
	if keyStatus > 0 then
		if actionName == 'MOUNTABLE_ATTACHER_MOUNT' then
			MountableAttacher.input_mountAttacher = true
		end
	end
	if MountableAttacher.controlledVehicle ~= nil then
		local vehicle = MountableAttacher.controlledVehicle
		local spec = vehicle.spec_mountableAttacher
		if actionName == 'MOUNTABLE_ATTACHER_MOVE_TRIGGER' then
			spec.input_moveTrigger = keyStatus > 0
		elseif actionName == 'MOUNTABLE_ATTACHER_MOVE_LEFT' then
			spec.input_moveLeft = keyStatus > 0
			if keyStatus == 0 then vehicle:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false) end
		elseif actionName == 'MOUNTABLE_ATTACHER_MOVE_RIGHT' then
			spec.input_moveRight = keyStatus > 0
			if keyStatus == 0 then vehicle:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false) end
		elseif actionName == 'MOUNTABLE_ATTACHER_MOVE_FORWARD' then
			spec.input_moveForward = keyStatus > 0
			if keyStatus == 0 then vehicle:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false) end
		elseif actionName == 'MOUNTABLE_ATTACHER_MOVE_BACKWARD' then
			spec.input_moveBackward = keyStatus > 0
			if keyStatus == 0 then vehicle:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false) end
		elseif actionName == 'MOUNTABLE_ATTACHER_MOVE_UP' then
			spec.input_moveUp = keyStatus > 0
			if keyStatus == 0 then vehicle:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false) end
		elseif actionName == 'MOUNTABLE_ATTACHER_MOVE_DOWN' then
			spec.input_moveDown = keyStatus > 0
			if keyStatus == 0 then vehicle:moveAttacher(spec.hookX, spec.hookY, spec.hookZ, false) end
		end
	end
end
function MountableAttacher:playerInRangeMA(target, dist)
	local spec = self.spec_mountableAttacher
	
	local inRange = false
	local source
	if g_currentMission.controlPlayer and g_currentMission.player ~= nil then
		source = g_currentMission.player.pickUpKinematicHelperNode
	end
	if source ~= nil and target ~= nil then
		local x,y,z = getWorldTranslation(target)
		local a,b,c = getWorldTranslation(source)
		local distance = MathUtil.vector3Length(x-a, y-b, z-c)
		if distance < dist then
			inRange = true
		end
	end
	return inRange
end
function MountableAttacher:onDraw()
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function MountableAttacher:maGetFullName()
    local name = self:getName()
    local storeItem = g_storeManager:getItemByXMLFilename(self.configFileName)
    if storeItem ~= nil then
        local brand = g_brandManager:getBrandByIndex(storeItem.brandIndex)
        if brand ~= nil then
            name = brand.title .. " " .. name
        end
    end
    return name
end

function MountableAttacher:onReadStream(streamId, connection)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	local attacherMounted = streamReadBool(streamId)
	
	if attacherMounted ~= nil and attacherMounted == true then
		local attacherMountedBack = streamReadBool(streamId)
		local hookX = streamReadFloat32(streamId)
		local hookY = streamReadFloat32(streamId)
		local hookZ = streamReadFloat32(streamId)
		self:mountAttacher(attacherMountedBack, true)
		self:moveAttacher(hookX, hookY, hookZ, true)
	end
end
function MountableAttacher:onWriteStream(streamId, connection)
	local spec = self.spec_mountableAttacher
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	streamWriteBool(streamId, spec.attacherMounted)
	if spec.attacherMounted then
		streamWriteBool(streamId, spec.attacherMountedBack)
		streamWriteFloat32(streamId, spec.hookX)
		streamWriteFloat32(streamId, spec.hookY)
		streamWriteFloat32(streamId, spec.hookZ)
	end
end

MountableAttacherMountEvent = {}
MountableAttacherMountEvent_mt = Class(MountableAttacherMountEvent, Event)
InitEventClass(MountableAttacherMountEvent, "MountableAttacherMountEvent")
function MountableAttacherMountEvent:emptyNew()
    local self = Event:new(MountableAttacherMountEvent_mt)
    self.className="MountableAttacherMountEvent"
    return self
end
function MountableAttacherMountEvent:new(object, attacherMountedBack)
	local self = MountableAttacherMountEvent:emptyNew()
	self.object = object
	self.attacherMountedBack = attacherMountedBack
	return self
end
function MountableAttacherMountEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.attacherMountedBack = streamReadBool(streamId)
    self:run(connection)
end
function MountableAttacherMountEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.attacherMountedBack)
end
function MountableAttacherMountEvent:run(connection)
	if self.object ~= nil then
		self.object:mountAttacher(self.attacherMountedBack, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(MountableAttacherMountEvent:new(self.object), nil, connection, self.object)
	end
end
function MountableAttacherMountEvent.sendEvent(vehicle, attacherMountedBack, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(MountableAttacherMountEvent:new(vehicle, attacherMountedBack), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(MountableAttacherMountEvent:new(vehicle, attacherMountedBack))
		end
	end
end

MountableAttacherUnmountEvent = {}
MountableAttacherUnmountEvent_mt = Class(MountableAttacherUnmountEvent, Event)
InitEventClass(MountableAttacherUnmountEvent, "MountableAttacherUnmountEvent")
function MountableAttacherUnmountEvent:emptyNew()
    local self = Event:new(MountableAttacherUnmountEvent_mt)
    self.className="MountableAttacherUnmountEvent"
    return self
end
function MountableAttacherUnmountEvent:new(object)
	local self = MountableAttacherUnmountEvent:emptyNew()
	self.object = object
	return self
end
function MountableAttacherUnmountEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
    self:run(connection)
end
function MountableAttacherUnmountEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
end
function MountableAttacherUnmountEvent:run(connection)
	if self.object ~= nil then
		self.object:unmountAttacher(true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(MountableAttacherUnmountEvent:new(self.object), nil, connection, self.object)
	end
end
function MountableAttacherUnmountEvent.sendEvent(vehicle, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(MountableAttacherUnmountEvent:new(vehicle), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(MountableAttacherUnmountEvent:new(vehicle))
		end
	end
end
MountableAttacherMoveEvent = {}
MountableAttacherMoveEvent_mt = Class(MountableAttacherMoveEvent, Event)

InitEventClass(MountableAttacherMoveEvent, "MountableAttacherMoveEvent")

function MountableAttacherMoveEvent:emptyNew()
    local self = Event:new(MountableAttacherMoveEvent_mt)
	self.className="MountableAttacherMoveEvent";
    return self;
end
function MountableAttacherMoveEvent:new(object, x, y, z)
    local self = MountableAttacherMoveEvent:emptyNew()
    self.object = object
    self.x = x
    self.y = y
    self.z = z
    return self
end
function MountableAttacherMoveEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId);
    self.x = streamReadFloat32(streamId)
    self.y = streamReadFloat32(streamId)
    self.z = streamReadFloat32(streamId)
    self:run(connection);
end
function MountableAttacherMoveEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object);
	streamWriteFloat32(streamId, self.x)
	streamWriteFloat32(streamId, self.y)
	streamWriteFloat32(streamId, self.z)
end
function MountableAttacherMoveEvent:run(connection)
	if self.object ~= nil then
		self.object:moveAttacher(self.x, self.y, self.z, true)
	end
    if not connection:getIsServer() then
		g_server:broadcastEvent(self, false, connection, self.object)
    end
end
function MountableAttacherMoveEvent.sendEvent(vehicle, x, y, z, noEventSend)
	if x ~= vehicle.x and y ~= vehicle.y and z ~= vehicle.z then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(MountableAttacherMoveEvent:new(vehicle, x, y, z), nil, nil, vehicle)
			else
				g_client:getServerConnection():sendEvent(MountableAttacherMoveEvent:new(vehicle, x, y, z))
			end
		end
	end
end